from functools import wraps
from copy import deepcopy
import hashlib
import sys
import inspect
import operator

from ..workflow import (from_call, get_workflow)
from .maybe import (maybe)
from ..lib import (decorator)
from noodles.config import config


def scheduled_function(f, hints=None):
    """The Noodles schedule function decorator.

    The decorated function will return a workflow in stead of
    being applied immediately. This workflow can then be passed to a job
    scheduler in order to be run on any architecture supporting the current
    python environment."""
    if hints is None:
        hints = {}

    if 'version' not in hints:
        try:
            source_bytes = inspect.getsource(f).encode()
        except Exception:
            pass
        else:
            m = hashlib.md5()
            m.update(source_bytes)
            hints['version'] = m.hexdigest()

    @wraps(f)
    def wrapped(*args, **kwargs):
        return PromisedObject(from_call(
            f, args, kwargs, deepcopy(hints),
            call_by_value=config['call_by_value']))

    # add *(scheduled)* to the beginning of the docstring.
    if hasattr(wrapped, '__doc__') and wrapped.__doc__ is not None:
        wrapped.__doc__ = "*(scheduled)* " + wrapped.__doc__

    return wrapped


@decorator
def schedule(f, **hints):
    """Decorator; schedule calls to function `f` into a workflow, in stead of
    running them at once. The decorated function returns a
    :class:`PromisedObject`."""
    return scheduled_function(f, hints=hints)


def has_scheduled_methods(cls):
    """Decorator; use this on a class for which some methods have been
    decorated with :func:`schedule` or :func:`schedule_hint`. Those methods
    are then tagged with the attribute `__member_of__`, so that we may
    serialise and retrieve the correct method. This should be considered
    a patch to a flaw in the Python object model."""
    for member in cls.__dict__.values():
        if hasattr(member, '__wrapped__'):
            member.__wrapped__.__member_of__ = cls

    return cls


def schedule_hint(**hints):
    """Decorator; same as :func:`schedule`, with added hints. These
    hints can be anything."""
    return lambda f: scheduled_function(f, hints)


@schedule
@maybe
def getitem(obj, ix):
    return obj[ix]


@schedule
def setitem(obj, attr, value):
    try:
        obj = deepcopy(obj)
        obj[attr] = value

    except TypeError as err:
        tb = sys.exc_info()[2]
        from pprint import PrettyPrinter
        pp = PrettyPrinter()
        obj_repr = "<" + obj.__class__.__name__ + ">: " \
            + pp.pformat(obj.__dict__)
        msg = "In `_setattr` we deepcopy the object *during runtime*. " \
              "If you're sure that what you're doing is safe, you can " \
              " overload `__deepcopy__` to get more efficient code. " \
              "However, something went " \
              "wrong here: \n" + err.args[0] + '\n' + obj_repr
        raise TypeError(msg).with_traceback(tb)

    return obj


@schedule
@maybe
def _getattr(obj, attr):
    return getattr(obj, attr)


@schedule
def _setattr(obj, attr, value):
    try:
        obj = deepcopy(obj)
        obj.__setattr__(attr, value)

    except TypeError as err:
        tb = sys.exc_info()[2]
        from pprint import PrettyPrinter
        pp = PrettyPrinter()
        obj_repr = "<" + obj.__class__.__name__ + ">: " \
            + pp.pformat(obj.__dict__)
        msg = "In `_setattr` we deepcopy the object *during runtime*. " \
              "If you're sure that what you're doing is safe, you can " \
              " overload `__deepcopy__` to get more efficient code. " \
              "However, something went " \
              "wrong here: \n" + err.args[0] + '\n' + obj_repr
        raise TypeError(msg).with_traceback(tb)

    return obj


@schedule
@maybe
def _do_call(obj, *args, **kwargs):
    return obj(*args, **kwargs)


def update_hints(obj, data):
    """Update the hints on the root-node of a workflow. Usually, schedule
    hints are fixed per function. Sometimes a user may want to set hints
    manually on a specific promised object. :func:`update_hints` uses the
    `update` method on the hints dictionary with `data` as its argument.

    :param obj: a :py:class:`PromisedObject`.
    :param data: a :py:class:`dict` containing additional hints.

    The hints are modified, in place, on the node. All workflows that contain
    the node are affected."""
    root = obj._workflow.root
    obj._workflow.nodes[root].hints.update(data)


def result(obj):
    """Results are stored on the nodes in the workflow at run time. This
    function can be used to get at a result of a node in a workflow after
    run time. This is not a recommended way of getting at results, but can
    help with debugging."""
    return obj.__result__()


class PromisedObject(object):
    """
    Wraps a :py:class:`Workflow`. The workflow represents the future promise
    of a Python object. By wrapping the workflow, we can mock the behaviour of
    this future object and schedule methods that were called by the user
    as if nothing weird is going on.
    """
    def __init__(self, workflow):
        self._workflow = workflow

    def __call__(self, *args, **kwargs):
        return _do_call(self._workflow, *args, **kwargs)

    def __getattr__(self, attr):
        return _getattr(self, attr)

    def __setattr__(self, attr, value):
        if attr[0] == '_':
            self.__dict__[attr] = value
            return

        self._workflow = get_workflow(
            _setattr(self, attr, value))

    def __result__(self):
        return self._workflow.root_node.result

    # predicates
    def __lt__(self, other):
        return schedule(operator.lt)(self, other)

    def __gt__(self, other):
        return schedule(operator.gt)(self, other)

    def __eq__(self, other):
        return schedule(operator.eq)(self, other)

    def __ne__(self, other):
        return schedule(operator.ne)(self, other)

    def __ge__(self, other):
        return schedule(operator.ge)(self, other)

    def __le__(self, other):
        return schedule(operator.le)(self, other)

    #  boolean operations
    def __bool__(self):
        return schedule(operator.truth)(self)

    # numerical operations
    def __abs__(self):
        return schedule(operator.abs)(self)

    def __sub__(self, other):
        return schedule(operator.sub)(self, other)

    def __add__(self, other):
        return schedule(operator.add)(self, other)

    def __mul__(self, other):
        return schedule(operator.mul)(self, other)

    def __rmul__(self, other):
        return schedule(operator.mul)(other, self)

    def __truediv__(self, other):
        return schedule(operator.truediv)(self, other)

    def __floordiv__(self, other):
        return schedule(operator.floordiv)(self, other)

    def __mod__(self, other):
        return schedule(operator.mod)(self, other)

    def __pow__(self, other):
        return schedule(operator.pow)(self, other)

    def __pos__(self):
        return schedule(operator.pos)(self)

    def __neg__(self):
        return schedule(operator.neg)(self)

    def __matmul__(self, other):
        return schedule(operator.matmul)(self, other)

    def __index__(self):
        return schedule(operator.index)(self)

    # bit operations
    def __inv__(self):
        return schedule(operator.inv)(self)

    def __lshift__(self, n):
        return schedule(operator.lshift)(self, n)

    def __rshift__(self, n):
        return schedule(operator.rshift)(self, n)

    def __and__(self, other):
        return schedule(operator.and_)(self, other)

    def __or__(self, other):
        return schedule(operator.or_)(self, other)

    def __xor__(self, other):
        return schedule(operator.xor)(self, other)

    # container operations
    def __contains__(self, item):
        return schedule(operator.contains)(self, item)

    def __getitem__(self, name):
        return getitem(self, name)

    def __setitem__(self, attr, value):
        self._workflow = get_workflow(
            setitem(self, attr, value))

    def __iter__(self):
        raise TypeError(
            "You have tried to iterate (or unpack) a PromisedObject. "
            "There is currently no possible way to learn the "
            "length of a PromisedObject so, sadly, this is not "
            "implemented. You may use the `noodles.unpack` function "
            "to unpack a promised tuple.")

    def __deepcopy__(self, _):
        raise TypeError(
            "A PromisedObject cannot be deepcopied. Most probably, you "
            "have a promise stored in another object, which you passed to "
            "a scheduled function. To transform an object with nested "
            "promises to a top-level promise, apply the ``lift`` function.")



import inspect
from itertools import count
from pathlib import Path
import base64

try:
    from dataclasses import is_dataclass  # noqa
    from .dataclass import SerDataClass
except ImportError:
    def is_dataclass(x):
        return False

    def SerDataClass():
        return None

from ..interface import (Fail, PromisedObject, Quote)
from ..lib import (object_name, look_up, importable, unwrap, is_unwrapped)
from ..workflow import (Workflow, NodeData, FunctionNode, ArgumentAddress,
                        ArgumentKind, reset_workflow, get_workflow)

from .registry import (Registry, Serialiser, SerUnknown)
from .reasonable import (Reasonable, SerReasonableObject)
from .path import (SerPath)


def ismethod(x):
    if inspect.ismethod(x):
        return True
    if type(x).__name__ == 'builtin_function_or_method' \
            and hasattr(x, '__self__'):
        return True
    return False


class SerAuto(Serialiser):
    def __init__(self):
        super(SerAuto, self).__init__('<automagic>')

    def encode(self, obj, make_rec):
        return obj.__serialize__(make_rec)

    def decode(self, cls, data):
        return cls.__construct__(data)


class SerByMembers(Serialiser):
    def __init__(self, cls, members):
        super().__init__(cls)
        self.members = members

    def encode(self, obj, make_rec):
        return make_rec({
            m: getattr(obj, m) for m in self.members
        })

    def decode(self, cls, data):
        return cls(**data)


class SerDict(Serialiser):
    def __init__(self):
        super(SerDict, self).__init__(dict)

    def encode(self, obj, make_rec):
        return make_rec(dict(obj))

    def decode(self, cls, data):
        return cls(data)


class SerBytes(Serialiser):
    def __init__(self):
        super(SerBytes, self).__init__(bytes)

    def encode(self, obj, make_rec):
        return make_rec(base64.b64encode(obj).decode())

    def decode(self, cls, data):
        return base64.b64decode(data.encode())


class SerSequence(Serialiser):
    """Tuples get converted to lists during serialisation.
    We want to get tuples back, so make this explicit."""
    def __init__(self, cls):
        super().__init__(cls)

    def encode(self, obj, make_rec):
        return make_rec(list(obj))

    def decode(self, cls, data):
        return cls(data)


class SerEnum(Serialiser):
    def __init__(self, cls):
        super(SerEnum, self).__init__(cls)

    def encode(self, obj, make_rec):
        return make_rec(obj.name)

    def decode(self, cls, data):
        return cls[data]


class SerNamedTuple(Serialiser):
    def __init__(self, cls):
        super(SerNamedTuple, self).__init__(cls)

    def encode(self, obj, make_rec):
        return make_rec(dict(obj._asdict()))

    def decode(self, cls, data):
        return cls(**data)


class SerSlice(Serialiser):
    def __init__(self):
        super(SerSlice, self).__init__(slice)

    def encode(self, obj, make_rec):
        return make_rec([obj.start, obj.stop, obj.step])

    def decode(self, cls, data):
        return slice(*data)


def _remap_links(remap, links):
    return [{'node': remap[source],
             'to': [{'node': remap[node],
                     'address': address}
                    for node, address in target]}
            for source, target in links.items()]


class SerWorkflow(Serialiser):
    def __init__(self):
        super(SerWorkflow, self).__init__(Workflow)

    def encode(self, obj, make_rec):
        remap = dict(zip(obj.nodes.keys(), count()))
        return make_rec({'root': remap[obj.root],
                         'nodes': list(obj.nodes.values()),
                         'links': _remap_links(remap, obj.links)})

    def decode(self, cls, data):
        root = data['root']

        nodes = dict(zip(count(), data['nodes']))

        links = {l['node']: {(target['node'], target['address'])
                             for target in l['to']}
                 for l in data['links']}

        return reset_workflow(Workflow(root, nodes, links))


class SerPromisedObject(Serialiser):
    def __init__(self):
        super(SerPromisedObject, self).__init__(PromisedObject)

    def encode(self, obj, make_rec):
        return make_rec({'workflow': get_workflow(obj)})

    def decode(self, cls, data):
        return PromisedObject(data['workflow'])


class SerMethod(Serialiser):
    def __init__(self):
        super(SerMethod, self).__init__('<method>')

    def encode(self, obj, make_rec):
        return make_rec({'class': object_name(obj.__member_of__),
                         'method': obj.__name__})

    def decode(self, cls, data):
        cls = look_up(data['class'])
        return unwrap(getattr(cls, data['method']))


class SerBoundMethod(Serialiser):
    def __init__(self):
        super(SerBoundMethod, self).__init__('<boundmethod>')

    def encode(self, obj, make_rec):
        return make_rec({
            'self': obj.__self__,
            'name': obj.__name__})

    def decode(self, _, data):
        return getattr(data['self'], data['name'])


class SerImportable(Serialiser):
    def __init__(self):
        super(SerImportable, self).__init__('<importable>')

    def encode(self, obj, make_rec):
        return make_rec(object_name(obj))

    def decode(self, cls, data):
        return look_up(data)


class SerNode(Serialiser):
    def __init__(self):
        super(SerNode, self).__init__(FunctionNode)

    def encode(self, obj, make_rec):
        return make_rec(dict(obj.data._asdict()))

    def decode(self, cls, data):
        return FunctionNode.from_node_data(NodeData(**data))


def _noodles_hook(obj):
    if '__member_of__' in dir(obj) and obj.__member_of__:
        return '<method>'

    if importable(obj):
        return '<importable>'

    if ismethod(obj):
        return '<boundmethod>'

    if is_unwrapped(obj):
        return '<unwrapped>'

    if hasattr(obj, '__serialize__') and hasattr(type(obj), '__construct__'):
        return '<automagic>'

    if is_dataclass(type(obj)) and not isinstance(obj, type):
        return '<dataclass>'

    return None


class SerUnwrapped(Serialiser):
    def __init__(self):
        return super().__init__('<unwrapped>')

    def encode(self, obj, make_rec):
        return make_rec(object_name(obj))

    def decode(self, cls, data):
        return unwrap(look_up(data))


def registry():
    """Returns the Noodles base serialisation registry."""
    return Registry(
        types={
            dict: SerDict(),
            tuple: SerSequence(tuple),
            set: SerSequence(set),
            bytes: SerBytes(),
            slice: SerSlice(),
            complex: SerByMembers(complex, ['real', 'imag']),
            Reasonable: SerReasonableObject(Reasonable),
            ArgumentKind: SerEnum(ArgumentKind),
            FunctionNode: SerNode(),
            ArgumentAddress: SerNamedTuple(ArgumentAddress),
            Workflow: SerWorkflow(),
            PromisedObject: SerPromisedObject(),
            Quote: SerReasonableObject(Quote),
            Path: SerPath(),
            Fail: SerReasonableObject(Fail)
        },
        hooks={
            '<method>': SerMethod(),
            '<boundmethod>': SerBoundMethod(),
            '<importable>': SerImportable(),
            '<automagic>': SerAuto(),
            '<unwrapped>': SerUnwrapped(),
            '<dataclass>': SerDataClass()
        },
        hook_fn=_noodles_hook,
        default=SerUnknown(),
    )



from ..worker import worker
from ..scheduler import Scheduler
from ...lib import (Queue, thread_pool)
from ...workflow import get_workflow

from itertools import repeat


def run_parallel(workflow, n_threads):
    """Run a workflow in parallel threads.

    :param workflow: Workflow or PromisedObject to evaluate.
    :param n_threads: number of threads to use (in addition to the scheduler).
    :returns: evaluated workflow.
    """
    scheduler = Scheduler()
    threaded_worker = Queue() >> thread_pool(
        *repeat(worker, n_threads))

    return scheduler.run(threaded_worker, get_workflow(workflow))



from .streams import (pull, patch)
from .queue import (Queue, EndOfQueue)
import threading
import functools


def thread_counter(finalize):
    """Modifies a thread target function, such that the number of active
    threads is counted. If the count reaches zero, a finalizer is called."""
    n_threads = 0
    lock = threading.Lock()

    def target_modifier(target):
        @functools.wraps(target)
        def modified_target(*args, **kwargs):
            nonlocal n_threads, lock

            with lock:
                n_threads += 1

            return_value = target(*args, **kwargs)

            with lock:
                n_threads -= 1
                if n_threads == 0:
                    finalize()

            return return_value

        return modified_target

    return target_modifier


def thread_pool(*workers, results=None, end_of_queue=EndOfQueue):
    """Returns a |pull| object, call it ``r``, starting a thread for each given
    worker.  Each thread pulls from the source that ``r`` is connected to, and
    the returned results are pushed to a |Queue|.  ``r`` yields from the other
    end of the same |Queue|.

    The target function for each thread is |patch|, which can be stopped by
    exhausting the source.

    If all threads have ended, the result queue receives end-of-queue.

    :param results: If results should go somewhere else than a newly
        constructed |Queue|, a different |Connection| object can be given.
    :type results: |Connection|

    :param end_of_queue: end-of-queue signal object passed on to the creation
        of the |Queue| object.

    :rtype: |pull|
    """
    if results is None:
        results = Queue(end_of_queue=end_of_queue)

    count = thread_counter(results.close)

    @pull
    def thread_pool_results(source):
        for worker in workers:
            t = threading.Thread(
                target=count(patch),
                args=(pull(source) >> worker, results.sink),
                daemon=True)
            t.start()

        yield from results.source()

    return thread_pool_results



"""
Manage IO between remote worker/pilot job, and the scheduler. Here there are
two options: use json, or msgpack.
"""

from ...lib import coroutine


def JSONObjectReader(registry, fi, deref=False):
    """Stream objects from a JSON file.

    :param registry: serialisation registry.
    :param fi: input file
    :param deref: flag, if True, objects will be dereferenced on decoding,
        otherwise we are lazy about decoding a JSON string.
    """
    for line in fi:
        yield registry.from_json(line, deref=deref)


@coroutine
def JSONObjectWriter(registry, fo, host=None):
    """Sink; writes object as JSON to a file.

    :param registry: serialisation registry.
    :param fo: output file.
    :param host: name of the host that encodes the JSON. This is relevant if
        the encoded data refers to external files for mass storage.

    In normal use, it may occur that the pipe to which we write is broken,
    for instance when the remote process shuts down. In that case, this
    coroutine exits.
    """
    while True:
        obj = yield
        try:
            print(registry.to_json(obj, host=host), file=fo, flush=True)
        except BrokenPipeError:
            return



import configparser

config = {
    # call_by_value: keep this True unless you know what you're doing.
    'call_by_value': True,
}


def read_config(filename: str):
    config = configparser.ConfigParser()
    config.read(filename)
    return config



class Connection(object):
    """Combine a source and a sink. These should represent the IO of
    some object, probably a worker. In this case the `source` is a
    coroutine generating results, while the sink needs to be fed jobs.

    .. |Connection| replace:: :py:class:`Connection`
    """
    def __init__(self, source, sink, aux=None):
        """Connection constructor

        :param source:
            The source signal coroutine
        :type source: generator

        :param sink:
            The signal sink coroutine
        :type sink: sink coroutine
        """
        self.source = source
        self.sink = sink
        self.aux = aux
        self.online = True

    def setup(self):
        """Activate the source and sink functions and return them in
        that order.

        :returns:
            source, sink
        :rtype: tuple"""
        return self.source(), self.sink()

    def __rshift__(self, other):
        return self.__join__(other)

    def __join__(self, other):
        """A connection has one output channel, so the '>>' operator
        connects the source to a coroutine, creating a new connection."""
        return Connection(self.source >> other, self.sink)



from .registry import Serialiser


class SerDataClass(Serialiser):
    def __init__(self):
        super(SerDataClass, self).__init__('<dataclass>')

    def encode(self, obj, make_rec):
        return make_rec(obj.__dict__)

    def decode(self, cls, data):
        return cls(**data)



from pathlib import Path
import uuid
import os
import sys

from noodles import serial
from ...lib import object_name


class WorkerConfig(object):
    """Configuration for a single remote Job.

    :param name:
        A quasi human recognizable name for this job. This will default to
        'remote-xxxxxx' where 'xxxxx' is some UUID.

    :param working_dir:
        The work directory where the job runs. Defaults to the current working
        directory.

    :param prefix:
        The path prefix of the Python system we're running on. This defaults to
        `sys.prefix`. If we are in a VirtualEnv, this means that the spawned
        job will run in the same VirtualEnv.

    :param exec_command:
        The command that is being executed. This defaults to `worker.sh` which
        should be located in the `working_dir`. This default script initialises
        the VirtualEnv en starts Python with `-m noodles.worker`, acting as a
        pilot job.

    :param init:
        You may specify a function that needs to be run before any jobs are
        being executed. There exist frameworks in the wild, which won't
        function otherwise.

    :param finish:
        This function may do some clean up, after all jobs have been done.

    :param verbose:
        Be verbose about what we're doing. This is for debugging purposes only.
    """
    def __init__(self, *, name=None, working_dir=None, prefix=None,
                 exec_command=None, n_threads=1, registry=serial.base,
                 init=None, finish=None, verbose=False):
        self.name = name or ("remote-" + str(uuid.uuid4()))
        self.working_dir = working_dir or os.getcwd()
        self.prefix = prefix or Path(sys.prefix)
        self.exec_command = exec_command
        self.n_threads = n_threads
        self.registry = registry
        self.init = init
        self.finish = finish
        self.verbose = verbose

    def command_line(self):
        executable = self.prefix / 'bin' / 'python'

        arguments = [
            '-m', 'noodles.pilot_job',
            '-name', self.name,
            '-registry', object_name(self.registry)]

        if self.init:
            arguments.extend(["-init", object_name(self.init)])

        if self.finish:
            arguments.extend(["-finish", object_name(self.finish)])

        if self.verbose:
            arguments.append("-verbose")

        return executable, arguments



from importlib import import_module


def object_name(obj):
    """Get the qualified name of an object. This will obtain both the module
    name from `__module__` and object name from `__name__`, and concatenate
    those with a '.'. Examples:

    >>> from math import sin
    >>> object_name(sin)
    'math.sin'

    >>> def f(x):
    ...     return x*x
    ...
    >>> object_name(f)
    '__main__.f'

    To have a qualified name, an object must be defined as a class or function
    in a module (``__main__`` is also a module). A normal instantiated object
    does not have a qualified name, even if it is defined and importable from a
    module.  Calling |object_name| on such an object will raise
    :py:exc:`AttributeError`.

    .. |object_name| replace:: :py:func:`object_name`"""
    return obj.__module__ + '.' + obj.__qualname__


def look_up(name):
    """Obtain an object from a qualified name. Example:

    >>> look_up('math.sin')
    <built-in function sin>

    This function should be considered the reverse of |object_name|.

    .. |look_up| replace:: :py:func:`look_up`"""
    path = name.split('.')
    module = import_module('.'.join(path[:-1]))
    return getattr(module, path[-1])


def importable(obj):
    """Check if an object can be serialised as a qualified name. This is done
    by checking that a ``look_up(object_name(obj))`` gives back the same
    object.

    .. |importable| replace:: :py:func:`importable`"""
    try:
        return look_up(object_name(obj)) is obj
    except (AttributeError, TypeError, ImportError):
        return False


def unzip_dict(d):
    a = {}
    b = {}

    for k, (v, w) in d.items():
        a[k] = v
        b[k] = w

    return a, b


def unwrap(f):
    """Safely obtain the inner function of a previously wrapped (or decorated)
    function. This either returns ``f.__wrapped__`` or just ``f`` if the latter
    fails.

    .. |unwrap| replace:: :py:func:`unwrap`"""
    try:
        return f.__wrapped__
    except AttributeError:
        return f


def is_unwrapped(f):
    """If `f` was imported and then unwrapped, this function might return True.

    .. |is_unwrapped| replace:: :py:func:`is_unwrapped`"""
    try:
        g = look_up(object_name(f))
        return g != f and unwrap(g) == f

    except (AttributeError, TypeError, ImportError):
        return False


def deep_map(f, root):
    """Sibling to |inverse_deep_map|. As :py:func:`map` maps over an iterable,
    |deep_map| maps over a structure of nested ``dict``s and ``list``s. Every
    object is passed through `f` recursively. That is, first `root` is mapped,
    next any object contained in its result, and so on.

    No distinction is made between tuples and lists. This function was
    created with encoding to JSON compatible data in mind.

    .. |deep_map| replace:: :py:func:`deep_map`"""
    result = f(root)

    if isinstance(result, dict):
        return {k: deep_map(f, v) for k, v in result.items()}

    if isinstance(result, list) or isinstance(result, tuple):
        return [deep_map(f, v) for v in result]

    return result


def inverse_deep_map(f, root):
    """Sibling to |deep_map|. Recursively maps objects in a nested structure of
    ``list`` and ``dict`` objects. Where |deep_map| starts at the top,
    |inverse_deep_map| starts at the bottom. First, if `root` is a ``list`` or
    ``dict``, its contents are |inverse_deep_map|ed. Then at the end, the
    entire object is passed through `f`.

    This function was created with decoding from JSON compatible data in mind.

    .. |inverse_deep_map| replace:: :py:func:`inverse_deep_map`"""
    if isinstance(root, dict):
        r = {k: inverse_deep_map(f, v) for k, v in root.items()}
    elif isinstance(root, list):
        r = [inverse_deep_map(f, v) for v in root]
    else:
        r = root

    return f(r)



import configparser
from typing import (List)
from ..lib import (look_up)


runners = [
    # ======================================================================= #
    #  Single threaded runners
    # ======================================================================= #
    {
        'name': 'single',
        'features': [],
        'description': """
            Run a workflow in a single thread. This is the absolute minimal
            runner, consisting of a single queue for jobs and a worker running
            jobs every time a result is pulled.""",
        'command': 'noodles.run.single',
        'arguments': {}
    },

    {
        'name': 'single',
        'features': ['display'],
        'description': """
            Adds a display to the single runner. Everything still runs in a
            single thread. Every time a job is pulled by the worker, a message
            goes to the display routine; when the job is finished the result is
            sent to the display routine.""",
        'command': 'noodles.run.single_with_display',
        'arguments': {
            'display': {
                'default': 'noodles.display.NCDisplay',
                'reader':  'look-up',
                'help': 'the display routine'
            }
         }
    },

    # ======================================================================= #
    #  Multi-threaded runners
    # ======================================================================= #
    {
        'name': 'parallel',
        'features': [],
        'command': 'noodles.run.parallel',
        'arguments': {
            'n_threads': {
                'default': '1',
                'reader': 'integer'
            }
        }
    },

    {
        'name': 'parallel',
        'features': ['display'],
        'command': 'noodles.run.parallel_with_display',
        'arguments': {
            'n_threads': {
                'default': '1',
                'reader': 'integer'
            },
            'display': {
                'default': 'noodles.display.NCDisplay',
                'reader':  'look-up'
            }
        }
    },

    {
        'name': 'parallel',
        'features': ['prov', 'display'],
        'description': """
            Run a workflow in `n_threads` parallel threads. Now we replaced the
            single worker with a thread-pool of workers.

            This version works with the JobDB to cache results; however we only
            store the jobs that are hinted with the 'store' keyword, unless
            `cache_all` is set to `True`.""",
        'command': 'noodles.run.run_with_prov.run_parallel_opt',
        'arguments': {
            'n_threads': {
                'default': '1',
                'reader': 'integer',
                'help': 'the number of threads to run'
            },
            'registry': {
                'default': 'noodles.serial.base',
                'reader': 'look-up',
                'help': 'the serialisation registry to use'
            },
            'display': {
                'default': 'noodles.display.NCDisplay',
                'reader':  'look-up',
                'help': 'the display to use'
            },
            'database': {
                'default': 'TinyDB',
                'help': 'the database backend for the job cache'
            },
            'cache_file': {
                'default': 'cache.json',
                'help': 'the file used to store the job cache'
            },
            'cache_all': {
                'default': 'False',
                'reader': 'boolean',
                'help': 'set this if you want to store all jobs in cache'
            }
        }
    },

    {
        'name': 'xenon',
        'features': ['prov', 'display'],
        'command': 'noodles.run.xenon.run_xenon_prov',
        'arguments': {
        }
    },

    {
        'name': 'process',
        'features': ['msgpack'],
        'command': 'noodles.run.process.run_process',
        'arguments': {
        }
    }
]


def find_runner(name: str, features: List[str]) -> str:
    name_candidates = filter(
        lambda r: r['name'] == name,
        runners)
    feature_candidates = filter(
        lambda r: all(f in r['features'] for f in features),
        name_candidates)
    return min(feature_candidates, lambda r: len(r['features']))


def run_with_config(config_file, workflow, machine=None):
    config = configparser.ConfigParser(
        interpolation=configparser.ExtendedInterpolation())
    config.read(config_file)

    machine = config.get('default', 'machine', fallback=machine)
    if machine is None:
        print("No machine given, running local in single thread.")
        runner = find_runner(name='single', features=[])
        settings = {}

    else:
        M = config['Machines']
        runner_name = M.get(machine, 'runner')
        features = map(str.strip, M.get(machine, 'features').split(','))
        runner = find_runner(name=runner_name, features=features)
        settings = dict(M[machine])

        del settings['runner']
        del settings['features']

        if 'user' in settings:
            settings['user'] = dict(config['Users'][settings['user']])

    run = look_up(runner['command'])

    return run(workflow, **settings)



from ..lib import (unwrap)
from .decorator import (
    PromisedObject, schedule, schedule_hint, has_scheduled_methods,
    update_hints, result)
from .functions import (
    delay, gather, gather_all, gather_dict, lift, unpack, quote,
    unquote, simple_lift, ref, Quote)
from .maybe import (maybe, Fail, failed)

__all__ = ['delay', 'gather', 'gather_all', 'gather_dict', 'schedule_hint',
           'schedule', 'unpack', 'has_scheduled_methods', 'unwrap',
           'update_hints', 'lift', 'failed',
           'PromisedObject', 'Quote', 'quote', 'unquote', 'result',
           'maybe', 'Fail', 'simple_lift', 'ref']




term_codes = {
    'fg':        (38, ';2;{0};{1};{2}'.format, 'm'),
    'bg':        (48, ';2;{0};{1};{2}'.format, 'm'),
    'bold':      (1,  'm'),
    'underline': (4,  'm'),
    'regular':   (23, 'm'),
    'italic':    (3,  'm'),
    'reset':     (0,  'm'),
    'back':      (lambda n=1: str(n), 'D'),
    'forward':   (lambda n=1: str(n), 'C'),
    'newline':   ('1E',),
    'up':        ('{0}'.format, 'A'),
    'down':      ('{0}'.format, 'B'),
    'save':      ('s',),
    'restore':   ('u',),
    'move':      ('{0};{1}'.format, 'H'),
    'clear':     ('2J',),
    'reverse':   ('7m',)}


def make_escape(cmd, *args):
    def f(x):
        if hasattr(x, '__call__'):
            return x(*args)
        else:
            return str(x)

    return '\033[' + ''.join(map(f, term_codes[cmd]))


class OutStream:
    def __init__(self, f):
        self.f = f

    def __lshift__(self, x):
        """Emulate the much liked C++ syntax for chaining output."""
        if isinstance(x, list):
            self.f.write(make_escape(*x))
            self.f.flush()
            return self

        if isinstance(x, str):
            self.f.write(x)
            self.f.flush()
            return self

        raise TypeError



"""
Implements parallel worker with Sqlite database.
"""

from itertools import repeat
import logging

from ..scheduler import (Scheduler)
from ..messages import (ResultMessage)
from ..worker import (worker)
from ..logging import make_logger

from ...workflow import (get_workflow)
from ...prov.sqlite import (JobDB)
from ...lib import (
    Queue, pull, thread_pool, Connection, EndOfQueue,
    pull_map, push_map)


def pass_job(db: JobDB, result_queue: Queue, always_cache=False):
    """Create a pull stream that receives jobs and passes them on to the
    database. If the job already has a result, that result is pushed onto
    the `result_queue`.
    """
    @pull
    def pass_job_stream(job_source):
        """Pull stream instance created by `pass_job`."""
        result_sink = result_queue.sink()

        for message in job_source():
            if message is EndOfQueue:
                return

            key, job = message
            if always_cache or ('store' in job.hints):
                status, retrieved_result = db.add_job_to_db(key, job)

                if status == 'retrieved':
                    result_sink.send(retrieved_result)
                    continue

                elif status == 'attached':
                    continue

            yield message

    return pass_job_stream


def pass_result(db: JobDB, always_cache=False):
    """Creates a pull stream receiving results, storing them in the database,
    then sending them on. At this stage, the database may return a list of
    attached jobs which also need to be sent on to the scheduler."""
    @pull
    def pass_result_stream(worker_source):
        """Pull stream instance created by `pass_result`."""
        for result in worker_source():
            if result is EndOfQueue:
                return

            attached = db.store_result_in_db(
                result, always_cache=always_cache)

            yield result
            yield from (ResultMessage(key, 'attached', result.value, None)
                        for key in attached)

    return pass_result_stream


def run_parallel(
        workflow, *, n_threads, registry, db_file, echo_log=True,
        always_cache=False):
    """Run a workflow in parallel threads, storing results in a Sqlite3
    database.

    :param workflow: Workflow or PromisedObject to evaluate.
    :param n_threads: number of threads to use (in addition to the scheduler).
    :param registry: serialization Registry function.
    :param db_file: filename of Sqlite3 database, give `':memory:'` to
        keep the database in memory only.
    :param echo_log: set log-level high enough
    :param always_cache: enable caching by schedule hint.
    :return: Evaluated result.
    """
    if echo_log:
        logging.getLogger('noodles').setLevel(logging.DEBUG)
        logging.debug("--- start log ---")

    with JobDB(db_file, registry) as db:
        job_queue = Queue()
        result_queue = Queue()

        job_logger = make_logger("worker", push_map, db)
        result_logger = make_logger("worker", pull_map, db)

        worker_pool = job_queue.source \
            >> pass_job(db, result_queue, always_cache) \
            >> thread_pool(*repeat(worker, n_threads), results=result_queue)
        job_front_end = job_logger >> job_queue.sink
        result_front_end = worker_pool \
            >> pass_result(db, always_cache) \
            >> result_logger

        scheduler = Scheduler(job_keeper=db)
        parallel_sqlite_worker = Connection(result_front_end, job_front_end)

        result = scheduler.run(parallel_sqlite_worker, get_workflow(workflow))

    return registry().dereference(result)



from noodles import (schedule, quote, unquote)
from typing import Any


def conditional(
        b: bool,
        branch_true: Any,
        branch_false: Any = None) -> Any:
    """
    Control statement to follow a branch
    in workflow. Equivalent to the `if` statement
    in standard Python.

    The quote function delay the evaluation of the branches
    until the boolean is evaluated.

    :param b:
       promised boolean value.
    :param branch_true:
       statement to execute in case of a true predicate.
    :param branch_false:
       default operation to execute in case of a false predicate.
    :returns: :py:class:`PromisedObject`
    """
    return schedule_branches(b, quote(branch_true), quote(branch_false))


@schedule
def schedule_branches(b: bool, quoted_true, quoted_false):
    """
    Helper function to choose which workflow to execute
    based on the boolean `b`.

    :param b:
       promised boolean value

    :param quoted_true:
       quoted workflow to eval if the boolean is true.
    :param quoted_true:
       quoted workflow to eval if the boolean is false.    """
    if b:
        return unquote(quoted_true)
    else:
        return unquote(quoted_false)



from ..lib import (pull_map, EndOfQueue)
from .messages import (ResultMessage, JobMessage)
import sys


@pull_map
def worker(job):
    """Primary |worker| coroutine. This is a |pull| object that pulls jobs from
    a source and yield evaluated results.

    Input should be of type |JobMessage|, output of type |ResultMessage|.

    .. |worker| replace:: :py:func::`worker`"""
    if job is EndOfQueue:
        return

    if not isinstance(job, JobMessage):
        print("Warning: Job should be communicated using `JobMessage`.",
              file=sys.stderr)

    key, node = job
    return run_job(key, node)


def run_job(key, node):
    """Run a job. This applies the function node, and returns a |ResultMessage|
    when complete. If an exception is raised in the job, the |ResultMessage|
    will have ``'error'`` status.

    .. |run_job| replace:: :py:func:`run_job`"""
    try:
        result = node.apply()
        return ResultMessage(key, 'done', result, None)

    except Exception as exc:
        return ResultMessage(key, 'error', None, exc)



"""
Provides logging facilities that can be inserted at any place in a system
of streams.
"""

import logging

from ..lib import (EndOfQueue)
from ..workflow import (is_workflow)
from .messages import (JobMessage, ResultMessage)


def _sugar(s):
    """Shorten strings that are too long for decency."""
    # s = s.replace("{", "{{").replace("}", "}}")
    if len(s) > 50:
        return s[:20] + " ... " + s[-20:]
    else:
        return s


def make_logger(name, stream_type, jobs):
    """Create a logger component.

    :param name: name of logger child, i.e. logger will be named
        `noodles.<name>`.
    :type name: str
    :param stream_type: type of the stream that this logger will
        be inserted into, should be |pull_map| or |push_map|.
    :type stream_type: function
    :param jobs: job-keeper instance.
    :type jobs: dict, |JobKeeper| or |JobDB|.

    :return: a stream.

    The resulting stream receives messages and sends them on after
    sending an INFO message to the logger. In the case of a |JobMessage|
    or |ResultMessage| a meaningful message is composed otherwise the
    string representation of the object is passed."""
    logger = logging.getLogger('noodles').getChild(name)
    # logger.setLevel(logging.DEBUG)

    @stream_type
    def log_message(message):
        if message is EndOfQueue:
            logger.info("-end-of-queue-")

        elif isinstance(message, JobMessage):
            logger.info(
                "job    %10s: %s", message.key, message.node)

        elif isinstance(message, ResultMessage):
            job = jobs[message.key]
            if is_workflow(message.value):
                logger.info(
                    "result %10s [%s]: %s -> workflow %x", message.key,
                    job.node, message.status, id(message.value))
            else:
                value_string = repr(message.value)
                logger.info(
                    "result %10s [%s]: %s -> %s", message.key, job.node,
                    message.status, _sugar(value_string))

        else:
            logger.info(
                "unknown message: %s", message)

        return message

    return log_message



"""
Implements single-threaded worker with Sqlite3 support.
"""

from ..scheduler import (Scheduler)
from ..messages import (ResultMessage)
from ..worker import (run_job)
from ..logging import make_logger

from ...workflow import (get_workflow)
from ...prov.sqlite import (JobDB)
from ...lib import (Queue, pull, pull_map, push_map, Connection)


def run_single(workflow, *, registry, db_file, always_cache=True):
    """"Run workflow in a single thread, storing results in a Sqlite3
    database.

    :param workflow: Workflow or PromisedObject to be evaluated.
    :param registry: serialization Registry function.
    :param db_file: filename of Sqlite3 database, give `':memory:'` to
        keep the database in memory only.
    :param always_cache: Currently ignored. always_cache is true.
    :return: Evaluated result.
    """
    with JobDB(db_file, registry) as db:
        job_logger = make_logger("worker", push_map, db)
        result_logger = make_logger("worker", pull_map, db)

        @pull
        def pass_job(source):
            """Receives jobs from source, passes back results."""
            for msg in source():
                key, job = msg
                status, retrieved_result = db.add_job_to_db(key, job)

                if status == 'retrieved':
                    yield retrieved_result
                    continue

                elif status == 'attached':
                    continue

                result = run_job(key, job)
                attached = db.store_result_in_db(result, always_cache=True)

                yield result
                yield from (ResultMessage(key, 'attached', result.value, None)
                            for key in attached)

        scheduler = Scheduler(job_keeper=db)
        queue = Queue()
        job_front_end = job_logger >> queue.sink
        result_front_end = queue.source >> pass_job >> result_logger
        single_worker = Connection(result_front_end, job_front_end)

        return scheduler.run(single_worker, get_workflow(workflow))



from .dynamic_pool import DynamicPool, xenon_interactive_worker
from ..scheduler import Scheduler
from ..job_keeper import JobKeeper

from ...workflow import (get_workflow)

from copy import copy


def run_xenon_simple(workflow, machine, worker_config):
    """Run a workflow using a single Xenon remote worker.

    :param workflow: |Workflow| or |PromisedObject| to evaluate.
    :param machine: |Machine| instance.
    :param worker_config: Configuration for the pilot job."""
    scheduler = Scheduler()

    return scheduler.run(
        xenon_interactive_worker(machine, worker_config),
        get_workflow(workflow)
    )


def run_xenon(
        workflow, *, machine, worker_config, n_processes, deref=False,
        verbose=False):
    """Run the workflow using a number of online Xenon workers.

    :param workflow: |Workflow| or |PromisedObject| to evaluate.
    :param machine: The |Machine| instance.
    :param worker_config: Configuration of the pilot job
    :param n_processes: Number of pilot jobs to start.
    :param deref: Set this to True to pass the result through one more
        encoding and decoding step with object dereferencing turned on.
    :returns: the result of evaluating the workflow
    """

    dynamic_pool = DynamicPool(machine)

    for i in range(n_processes):
        cfg = copy(worker_config)
        cfg.name = 'xenon-{0:02}'.format(i)
        dynamic_pool.add_xenon_worker(cfg)

    job_keeper = JobKeeper()
    S = Scheduler(job_keeper=job_keeper, verbose=verbose)

    result = S.run(
        dynamic_pool, get_workflow(workflow)
    )

    dynamic_pool.close_all()

    if deref:
        return worker_config.registry().dereference(result, host='scheduler')
    else:
        return result



from .coroutine import coroutine
import inspect


class stream(object):
    """Base class for *pull* and *push* coroutines."""
    def __rshift__(self, other):
        return self.__join__(other)


class pull(stream):
    """A |pull| coroutine pulls from a source, yielding values.
    |pull| Objects can be chained using the `>>` operator.

    A |pull| object acts as a function of one argument, being the
    source that the coroutine will pull from. This source argument
    must always be a thunk (function of zero arguments) returning
    an iterable.

    .. |pull| replace:: :py:class:`pull`
    """
    def __init__(self, fn):
        self.source = fn

    def __iter__(self):
        return self.source()

    def __join__(self, other):
        if isinstance(other, pull):
            return pull(lambda *x: other.source(lambda: self.source(*x)))
        elif inspect.isfunction(other):
            return self >> pull_map(other)
        else:
            raise TypeError(
                'Don\'t know how to join {} and {}'.format(self, other))

    def __call__(self, *x):
        return self.source(*x)


class push(stream):
    """A |push| coroutine pushes to a sink, receiving values through ``yield``
    statements. |push| Objects can be chained using the `>>` operator.

    A |push| object acts as a function of one argument, being the sink that the
    coroutine will send to. This sink argument must always be a thunk (function
    of zero arguments) returning a coroutine.

    .. |push| replace:: :py:class:`push`
    """
    def __init__(self, fn, dont_wrap=False):
        if dont_wrap:
            self.sink = fn
        else:
            self.sink = coroutine(fn)

    def __lshift__(self, other):
        return other.__join__(self)

    def __join__(self, other):
        if isinstance(other, push):
            return push(lambda *x: self.sink(lambda: other.sink(*x)),
                        dont_wrap=True)
        elif inspect.isfunction(other):
            return self >> push_map(other)
        else:
            raise TypeError(
                'Don\'t know how to join {} and {}'.format(self, other))

    def __call__(self, *x):
        return self.sink(*x)


def compose_2(f, g):
    def h(*x):
        return f(g(*x))

    return h


class pull_map(pull):
    """A |pull_map| decorates a function of a single argument, to become
    a |pull| object. The resulting |pull| object pulls object from a source
    yielding values mapped through the given function.

    This is equivalent to::

        @pull
        def g(source):
            yield from map(f, source())

    where ``f`` is the function being decorated.

    The ``>>`` operator on this class is optimised such that only a single
    loop will be created when chained with another |pull_map|.

    Also, a |pull_map| may be chained to a function directly, including
    the given function in the loop.

    .. |pull_map| replace:: :py:class:`pull_map`
    """
    def __init__(self, f):
        self.f = f
        super(pull_map, self).__init__(self.map)

    def __join__(self, other):
        if not isinstance(other, stream) and inspect.isfunction(other):
            return pull_map(compose_2(other, self.f))

        if isinstance(other, pull_map):
            return pull_map(compose_2(other.f, self.f))

        else:
            return super(pull_map, self).__join__(other)

    def map(self, source):
        yield from map(self.f, source())


class push_map(push):
    """A |push_map| decorates a function of a single argument, to become
    a |push| object. The resulting |push| object receives values through
    ``yield`` and sends them on after mapping through the given function.

    This is equivalent to::

        @push
        def g(sink):
            sink = sink()
            while True:
                x = yield
                sink.send(f(x))

    where ``f`` is the function being decorated.

    The ``>>`` operator on this class is optimised such that only a single
    loop will be created when chained with another |push_map|.

    Also, a |push_map| may be chained to a function directly, including
    the given function in the loop.

    .. |push_map| replace:: :py:class:`push_map`
    """
    def __init__(self, f):
        self.f = f
        super(push_map, self).__init__(self.map)

    def __join__(self, other):
        if not isinstance(other, stream) and inspect.isfunction(other):
            return push_map(compose_2(other, self.f))

        if isinstance(other, push_map):
            return push_map(compose_2(other.f, self.f))

        else:
            return super(push_map, self).__join__(other)

    def map(self, sink):
        sink = sink()

        while True:
            try:
                x = yield
                sink.send(self.f(x))
            except StopIteration:
                return


def sink_map(f):
    """The |sink_map| decorator creates a |push| object from a function that
    returns no values. The resulting sink can only be used as an end point of a
    chain.

    Equivalent code::

        @push
        def sink():
            while True:
                x = yield
                f(x)

    .. |sink_map| replace:: :py:func:`sink_map`
    """
    @push
    def sink():
        while True:
            args = yield
            f(args)

    return sink


def branch(*sinks_):
    """The |branch| decorator creates a |pull| object that pulls from a single
    source and then sends to all the sinks given. After all the sinks received
    the message, it is yielded.

    .. |branch| replace:: :py:func:`branch`
    """
    @pull
    def junction(source):
        sinks = [s() for s in sinks_]

        for msg in source():
            for s in sinks:
                s.send(msg)
            yield msg

    return junction


def broadcast(*sinks_):
    """The |broadcast| decorator creates a |push| object that receives a
    message by ``yield`` and then sends this message on to all the given sinks.

    .. |broadcast| replace:: :py:func:`broadcast`
    """
    @push
    def bc():
        sinks = [s() for s in sinks_]
        while True:
            msg = yield
            for s in sinks:
                s.send(msg)

    return bc


def pull_from(iterable):
    """Creates a |pull| object from an iterable.

    :param iterable: an iterable object.
    :type iterable: :py:class:`~collections.abc.Iterable`
    :rtype: |pull|

    Equivalent to::

        pull(lambda: iter(iterable))
    """
    return pull(lambda: iterable)


def push_from(iterable):
    """Creates a |push| object from an iterable. The resulting function
    is not a coroutine, but can be chained to another |push|.

    :param iterable: an iterable object.
    :type iterable: :py:class:`~collections.abc.Iterable`
    :rtype: |push|
    """
    def p(sink):
        sink = sink()
        for x in iterable:
            sink.send(x)

    return push(p, dont_wrap=True)


def patch(source, sink):
    """Create a direct link between a source and a sink.

    Implementation::

        sink = sink()
        for value in source():
            sink.send(value)

    .. |patch| replace:: :py:func:`patch`
    """
    sink = sink()
    for v in source():
        try:
            sink.send(v)
        except StopIteration:
            return



try:
    from .numpy import registry as numpy
except ImportError:
    numpy = None

from .pickle import registry as pickle
from .base import registry as base
from .as_dict import AsDict
from .namedtuple import registry as namedtuple
from .registry import (Registry, Serialiser, RefObject)
from .reasonable import (Reasonable)
from .path import SerPath

__all__ = ['pickle', 'base', 'Registry', 'Serialiser', 'SerPath',
           'RefObject', 'AsDict', 'namedtuple', 'Reasonable', 'numpy']



import sys
import os
import curses
import termios


class termcap(dict):
    """This is a dictionary class that stores all required termios capabilities
    for you. Just initialize it with a dictionary containing three list of
    strings. A list for boolean (or flag caps) called 'flag', a 'num' list for
    numerical caps, and a 'str' for strings. Or just leave it empty, and add
    the items manually later on."""
    def __init__(self, cap_list):
        dict.__init__(self)

        for i in cap_list['flag']:
            self.getflag(i)

        for i in cap_list['num']:
            self.getnum(i)

        for i in cap_list['str']:
            self.getstr(i)

    def getstr(self, capname):
        self[capname] = curses.tigetstr(capname)
        if self[capname] is None:
            raise RuntimeError(
                'This terminal is not capable of doing ' + capname)

    def getnum(self, capname):
        self[capname] = curses.tigetnum(capname)
        if self[capname] is None:
            raise RuntimeError(
                'This terminal is not capable of doing ' + capname)

    def getflag(self, capname):
        self[capname] = curses.tigetflag(capname)
        if self[capname] is None:
            raise RuntimeError(
                'This terminal is not capable of doing ' + capname)


class ostream:
    def __init__(self, fo, caps, extra=None):
        self.fo = fo
        self.caps = caps
        self.extra = extra

    def __lshift__(self, q):
        if isinstance(q, str):
            self.fo.write(q)

        elif isinstance(q, list):
            cmd = q[0]
            args = q[1:]

            if cmd in self.caps:
                self.fo.write(curses.tparm(
                    self.caps[cmd], *args).decode('ascii'))
            elif self.extra and cmd in self.extra:
                self.fo.write(self.extra[cmd](*args))
            else:
                raise RuntimeError('Could not find terminfo code: ' + cmd)

        elif hasattr(q, '__iter__'):
            for item in q:
                self << item

        self.fo.flush()
        return self


class TermApp(ostream):
    required_caps = {
        'flag': [
            # mainly for my debugging pursoses, as the currents set of caps
            # should work for all xterm and rxvt derivates
            'bce', 'ccc', 'cpix', 'crxm', 'km', 'hs', 'lpix', 'da',
            'db', 'mir', 'ndscr', 'nrrmc', 'sam', 'eslok', 'xon',
        ],
        'num': [
            'cols',        # number of columns
            'lines',    # number of lines
            'colors',    # number of colors, preferably 8.
        ],
        'str': [
            'cup',        # move cursor to #1, #2
            'cud',        # cursor down
            'cuu',        # cursor up
            'cub',        # cursor back
            'cuf',        # cursor forward
            'cr',        # cariage return
            'sc',        # save cursor
            'rc',        # restore cursor

            'civis',    # invisible cursor
            'cnorm',    # visible cursor

            'clear',    # clear the entire screen
            'el1',        # clear to beginning of line
            'el',        # clear to end of line
            'ed',        # clear to end of display
            'dch',        # clear n chars
            'dl',        # delete n lines

            'op',        # reset terminal colors
            'sgr0',        # turn off attributes
            'sitm',     # italics
            'bold',        # bold mode
            'rev',        # reverse mode
            'smul',        # underline mode
            'setab',    # set background
            'setaf'    # set foreground
            # 'setb',    # set background
            # 'setf',    # set foreground
        ]
    }

    def __init__(self):
        self.input = sys.stdin
        self.output = sys.stdout

        curses.setupterm(os.getenv("TERM"), self.output.fileno())
        caps = termcap(self.required_caps)
        extra = {
            'setfg': '\033[38;2;{0};{1};{2}m'.format,
            'setbg': '\033[48;2;{0};{1};{2}m'.format
        }

        super(TermApp, self).__init__(self.output, caps, extra)

    def __enter__(self):
        curses.setupterm()
        self.orig_attr = termios.tcgetattr(self.output.fileno())

        self.new_attr = list(self.orig_attr)  # make a copy
        self.new_attr[3] &= ~(termios.ECHO | termios.ICANON)
        self.new_attr[6][termios.VMIN] = 0
        self.new_attr[6][termios.VTIME] = 0
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.restore_term()

    def restore_term(self):
        self << "\033[?25h"  # cursor on
        termios.tcsetattr(self.output.fileno(), termios.TCSANOW,
                          self.orig_attr)

    def setup_term(self):
        self << "\033[?25l"  # cursor off
        termios.tcsetattr(self.output.fileno(), termios.TCSANOW,
                          self.new_attr)

    def set_im(self, vt, vm):
        self.new_termios[6][termios.VTIME] = vt
        self.new_termios[6][termios.VMIN] = vm
        termios.tcsetattr(self.output, termios.TCSANOW, self.new_termios)

    def refresh_size(self):
        self.getnum('lines')
        self.getnum('cols')

    def get_cols(self):
        return self.caps['cols']

    def get_lines(self):
        return self.caps['lines']

    def clear(self):
        return self.caps['clear']

    def set_fg(self, c):
        return curses.tparm(self.caps['setaf'], c)

    def set_bg(self, c):
        return curses.tparm(self.caps['setab'], c)

    def set_bold(self):
        return self.caps['bold']

    def set_rv(self):
        return self.caps['rev']

    def set_ul(self):
        return self.caps['smul']

    def moveyx(self, y, x):
        return curses.tparm(self.caps['cup'], y, x)

    def movexy(self, x, y):
        return self.moveyx(y, x)

    def move_down(self, n=1):
        return curses.tparm(self.caps['cud'], n)

    def move_up(self, n=1):
        return curses.tparm(self.caps['cuu'], n)

    def move_left(self, n=1):
        return curses.tparm(self.caps['cub'], n)

    def move_right(self, n=1):
        return curses.tparm(self.caps['cuf'], n)

    def cr(self):
        return self.caps['cr']

    def save_pos(self):
        return self.caps['sc']

    def restore_pos(self):
        return self.caps['rc']

    def cursor_on(self):
        return self.caps['cnorm']

    def cursor_off(self):
        return self.caps['civis']

    def reset_color(self):
        return self.caps['op'] + self.caps['sgr0']



from .pretty_term import OutStream
from ..workflow import FunctionNode
from inspect import Parameter
import sys


def _format_arg_list(a, v):
    if len(a) == 0:
        if v:
            return "(\u2026)"
        else:
            return "()"

    s = "({0}{1})"
    for i in a[:-1]:
        s = s.format(str(i) if i != Parameter.empty else "\u2014", ", {0}{1}")

    if v:
        return s.format("\u2026", "")

    return s.format(str(a[-1]) if a[-1] != Parameter.empty else "\u2014", "")


class DumbDisplay:
    """Monochrome, dumb term display"""
    def __init__(self, error_filter=None):
        self.jobs = {}
        self.out = OutStream(sys.stdout)
        self.errors = []
        self.error_filter = error_filter
        self.messages = []

    def print_message(self, key, msg):
        if key in self.jobs:
            print("{1:12} | {2}".format(
                    key, '['+msg.upper()+']', self.jobs[key]['name']),
                  file=sys.stderr)

    def add_job(self, key, name):
        self.jobs[key] = {'name': name}

    def error_handler(self, job, xcptn):
        self.errors.append((job, xcptn))

    def report(self):
        if len(self.errors) == 0:
            self.out << "[success]\n"

        else:
            self.out << "[ERROR!]\n\n"

            for job, e in self.errors:
                msg = 'ERROR '
                if 'display' in job.hints:
                    msg += job.hints['display'].format(
                        **job.bound_args.arguments)
                else:
                    msg += 'calling {} with {}'.format(
                        job.foo.__name__, dict(job.bound_args.arguments)
                    )

                print(msg)
                err_msg = self.error_filter(e)
                if err_msg:
                    print(err_msg)
                else:
                    print(e)

    def __call__(self, msg):
        key, status, data, err = msg

        if isinstance(data, FunctionNode) and hasattr(data, 'hints'):
            job = data
            if job.hints and 'display' in job.hints:
                msg = job.hints['display'].format(**job.bound_args.arguments)
            else:
                msg = "{0} {1}".format(
                    job.foo.__name__,
                    _format_arg_list(job.bound_args.args, None))

            self.add_job(key, msg)

        if hasattr(self, status):
            getattr(self, status)(key, data, err)
        else:
            self.print_message(key, status)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # self.wait()

        if exc_type:
            if exc_type is KeyboardInterrupt:
                self.out << "\n" << "User interrupt detected, abnormal exit.\n"
                return True

            print("Internal error encountered. Contact the developers.")
            return False

        self.report()

    def wait(self):
        self.q.wait()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['noodles',
 'noodles.display',
 'noodles.display.curses',
 'noodles.draw_workflow',
 'noodles.interface',
 'noodles.lib',
 'noodles.patterns',
 'noodles.prov',
 'noodles.run',
 'noodles.run.remote',
 'noodles.run.single',
 'noodles.run.threading',
 'noodles.run.xenon',
 'noodles.serial',
 'noodles.workflow']

package_data = \
{'': ['*']}

install_requires = \
['filelock>=3.4.2,<4.0.0',
 'graphviz>=0.19.1,<0.20.0',
 'h5py>=3.6.0,<4.0.0',
 'numpy>=1.22.2,<2.0.0',
 'pyxenon>=3.0.3,<4.0.0']

setup_kwargs = {
    'name': 'noodles',
    'version': '0.3.4',
    'description': 'Worflow Engine',
    'long_description': None,
    'author': 'Johan Hidding',
    'author_email': 'j.hidding@esciencecenter.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
.. |ArgumentKind| replace:: :py:class:`ArgumentKind`
.. |ArgumentAddress| replace:: :py:class:`ArgumentAddress`
.. |Argument| replace:: :py:class:`Argument`
"""

from collections import namedtuple
from enum import Enum
from itertools import repeat
from inspect import Parameter, signature

Empty = Parameter.empty

"""Denotes the kind of argument."""
ArgumentKind = Enum(
    'ArgumentKind',
    ['regular', 'variadic', 'keyword'])

"""Codifies the location to a unique argument of a function."""
ArgumentAddress = namedtuple(
    'ArgumentAddress',
    ['kind', 'name', 'key'])

"""Codifies a value given for some argument."""
Argument = namedtuple(
    'Argument',
    ['address', 'value'])


def serialize_arguments(bound_args):
    """Generator that takes the bound_args output of signature().bind and
    iterates over all the arguments, returning reproducable addresses of each
    argument.

    An address is stored in an |ArgumentAddress| object (a named tuple),
    containing the kind of argument (regular, variadic or keyword), the name of
    the argument, and, if not a regular argument, a key.  In the case of a
    variadic argument this is an integer index into the variadic arguments
    list, in the case of a keyword argument it is a string. For regular
    arguments the key is set to `None`.

    :param bound_args:
        Bound arguments structure, as described in the documentation of the
        `inspect` module.
    :type bound_args: BoundArguments

    :returns:
        Generates (kind, name, key)-tuples representing an address into the
        argument structure.
    :rtype: Iterator[ArgumentAddress]
    """
    for p in bound_args.signature.parameters.values():
        if p.kind == Parameter.VAR_POSITIONAL:
            for i, _ in enumerate(bound_args.arguments[p.name]):
                yield ArgumentAddress(ArgumentKind.variadic, p.name, i)
            continue

        if p.kind == Parameter.VAR_KEYWORD:
            for k in bound_args.arguments[p.name].keys():
                yield ArgumentAddress(ArgumentKind.keyword, p.name, k)
            continue

        yield ArgumentAddress(ArgumentKind.regular, p.name, None)


def ref_argument(bound_args, address):
    """ Taking a bound_args object, and an ArgumentAddress, retrieves the data
    currently stored in bound_args for this particular address."""
    if address.kind == ArgumentKind.regular:
        return bound_args.arguments[address.name]

    return bound_args.arguments[address.name][address.key]


def set_argument(bound_args, address, value):
    """ Taking a bound_args object, and an |ArgumentAddress| and a value, sets
    the value pointed to by the address to `value`."""
    if address.kind == ArgumentKind.regular:
        bound_args.arguments[address.name] = value
        return

    if address.kind == ArgumentKind.variadic:
        if address.name not in bound_args.arguments:
            bound_args.arguments[address.name] = []

        n_args = len(bound_args.arguments[address.name])
        if address.key >= n_args:
            bound_args.arguments[address.name].extend(
                repeat(Empty, address.key - n_args + 1))

    if address.kind == ArgumentKind.keyword:
        if address.name not in bound_args.arguments:
            bound_args.arguments[address.name] = {}

    bound_args.arguments[address.name][address.key] = value


def format_address(address):
    """Formats an ArgumentAddress for human reading."""
    if address.kind == ArgumentKind.regular:
        return address.name

    return "{0}[{1}]".format(address.name, address.key)


def get_arguments(bound_args):
    return [(address, ref_argument(bound_args, address))
            for address in serialize_arguments(bound_args)
            if ref_argument(bound_args, address) is not Empty]


def bind_arguments(f, arguments):
    bound_args = signature(f).bind_partial()

    variadic = next((x.name
                     for x in bound_args.signature.parameters.values()
                     if x.kind == Parameter.VAR_POSITIONAL), None)

    if variadic:
        bound_args.arguments[variadic] = []

    keyword = next((x.name
                    for x in bound_args.signature.parameters.values()
                    if x.kind == Parameter.VAR_KEYWORD), None)

    if keyword:
        bound_args.arguments[keyword] = {}

    for address, value in arguments:
        set_argument(bound_args, address, value)

    return bound_args



import threading

from ..workflow import get_workflow
from ..lib import Queue, Connection, push, patch, EndOfQueue, FlushQueue
from .scheduler import Scheduler
from .worker import run_job


def hybrid_coroutine_worker(selector, workers):
    """Runs a set of workers, all of them in the main thread.
    This runner is here for testing purposes.

    :param selector:
        A function returning a worker key, given a job.
    :type selector: function

    :param workers:
        A dict of workers.
    :type workers: dict
    """
    jobs = Queue()

    worker_source = {}
    worker_sink = {}

    for k, w in workers.items():
        worker_source[k], worker_sink[k] = w.setup()

    def get_result():
        source = jobs.source()

        for msg in source:
            key, job = msg
            worker = selector(job)
            if worker is None:
                yield run_job(key, job)
            else:
                # send the worker a job and wait for it to return
                worker_sink[worker].send(msg)
                result = next(worker_source[worker])
                yield result

    return Connection(get_result, jobs.sink)


def hybrid_threaded_worker(selector, workers):
    """Runs a set of workers, each in a separate thread.

    :param selector:
        A function that takes a hints-tuple and returns a key
        indexing a worker in the `workers` dictionary.
    :param workers:
        A dictionary of workers.

    :returns:
        A connection for the scheduler.
    :rtype: Connection

    The hybrid worker dispatches jobs to the different workers
    based on the information contained in the hints. If no hints
    were given, the job is run in the main thread.

    Dispatching is done in the main thread. Retrieving results is
    done in a separate thread for each worker. In this design it is
    assumed that dispatching a job takes little time, while waiting for
    one to return a result may take a long time.
    """
    result_queue = Queue()

    job_sink = {k: w.sink() for k, w in workers.items()}

    @push
    def dispatch_job():
        default_sink = result_queue.sink()

        while True:
            msg = yield

            if msg is EndOfQueue:
                for k in workers.keys():
                    try:
                        job_sink[k].send(EndOfQueue)
                    except StopIteration:
                        pass
                return

            if msg is FlushQueue:
                for k in workers.keys():
                    try:
                        job_sink[k].send(FlushQueue)
                    except StopIteration:
                        pass
                return

            worker = selector(msg.node)
            if worker:
                job_sink[worker].send(msg)
            else:
                default_sink.send(run_job(*msg))

    for key, worker in workers.items():
        t = threading.Thread(
            target=patch,
            args=(worker.source, result_queue.sink))
        t.daemon = True
        t.start()

    return Connection(result_queue.source, dispatch_job)


def run_hybrid(wf, selector, workers):
    """
    Returns the result of evaluating the workflow; runs through several
    supplied workers in as many threads.

    :param wf:
        Workflow to compute
    :type wf: :py:class:`Workflow` or :py:class:`PromisedObject`

    :param selector:
        A function selecting the worker that should be run, given a hint.
    :param workers:
        A dictionary of workers

    :returns:
        result of running the workflow
    """
    worker = hybrid_threaded_worker(selector, workers)
    return Scheduler().run(worker, get_workflow(wf))



import functools


def decorator(f):
    """Creates a paramatric decorator from a function. The resulting decorator
    will optionally take keyword arguments."""
    @functools.wraps(f)
    def decoratored_function(*args, **kwargs):
        if args and len(args) == 1:
            return f(*args, **kwargs)

        if args:
            raise TypeError(
                "This decorator only accepts extra keyword arguments.")

        return lambda g: f(g, **kwargs)

    return decoratored_function



from pathlib import Path
from .registry import (Serialiser)


class SerPath(Serialiser):
    def __init__(self):
        super(SerPath, self).__init__(Path)

    def encode(self, obj, make_rec):
        return make_rec(str(obj), files=[str(obj)])

    def decode(self, cls, data):
        return Path(data)
